package graphics.grapher;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.net.URL;
import java.util.Vector;


public class GraphComponent extends JComponent {

   private Image imageData = null;

/*
**  The axis we are controlling
*/
    private Vector axis = new Vector(4);
/*
**  The datasets we are controlling
*/
    private Vector dataset = new Vector(10);
/*
**  The markers that have been loaded
*/
    private Markers marker;

/*
** The blinking "data loading" thread
*/
    private LoadMessage load_thread = null;

    /**
     * ****************** * * Public Variables * *******************
     */

/*
**  loadingData
**              If this is greater than zero it means that
**  data loading threads are active so the message "loading data"
**  is flashed on the plot canvas. When it is back to zero the plot
**  progresses normally
*/
    private int loadingData = 0;

/*
**  The border around the entire plot. Allows slopover from axis labels
**  legends etc.
*/
    private int borderTop = 10;
    private int borderBottom = 10;
    private int borderLeft = 10;
    private int borderRight = 10;
/*
**  Do you want a frame around the data window. Axis will overlay this frame
**  Not forgetting the frame color
*/
    private boolean frame = true;

/*
**  Do you want a grid over the data window.
*/
    private boolean drawgrid = true;
    private Color gridcolor = Color.pink;
/*
**  Do you want grid lines along the zero's
*/
    private boolean drawzero = true;
    private Color zerocolor = Color.orange;
/*
**  The rectangle where the data will be plotted
*/
    private Rectangle datarect = new Rectangle();
/*
**  Blank the screen when the update method is called? Normally
**  only modified for special effects
*/
    private boolean clearAll = true;
/*
**  Paint it All ie axis, data etc. Normally
**  only modified for special effects
*/
    private boolean paintAll = true;


    /**
     * *************** * *  Public Methods * *****************
     */


/*
**  loadDataSet( double data[], int n )  
**
**  Load a DataSet from an Array. The data is assumed to be stored
**  in the form  x,y,x,y,x,y.... A local copy of the data is made.
**
**  If everything goes well it returns 'true'
**
**  double data[]: The data to plot stored in the form x,y,x,y...
**  int   n:      The number of data points. This means that the
**                minimum length of the data array is 2*n.
*/

    public DataSet loadDataSet(double data[], int n) {
        DataSet d;
        try {
            d = new DataSet(data, n);
            dataset.addElement(d);
            d.g2d = this;
        } catch (Exception e) {
            System.out.println("Failed to load Data set ");
            e.printStackTrace();
            return null;
        }
        return d;
    }

/*
**  attachDataSet( DataSet d )  
**      Add a data set to the graph
*/

    public void attachDataSet(DataSet d) {

        if (d != null) {
            dataset.addElement(d);
            d.g2d = this;
        }
    }
/*
** deleteDataSet( DataSet d ) 
**        Delete all pointers to the local copy for the data set.
*/

    public void deleteDataSet(DataSet d) {
        if (d != null) {
            if (d.xaxis != null) d.xaxis.detachDataSet(d);
            if (d.yaxis != null) d.yaxis.detachDataSet(d);
            dataset.removeElement(d);
        }
    }
/*
** createAxis( int position )
**            create an axis
*/
    public Axis createAxis(int position) {
        Axis a;

        try {
            a = new Axis(position);
            axis.addElement(a);
        } catch (Exception e) {
            System.out.println("Failed to create Axis");
            e.printStackTrace();
            return null;
        }

        return a;
    }
/*
** attachAxis( Axis a )
**           attach an existing axis
*/
    public void attachAxis(Axis a) {

        if (a == null) return;

        try {
            axis.addElement(a);
        } catch (Exception e) {
            System.out.println("Failed to attach Axis");
            e.printStackTrace();
        }

    }

/*
** deleteAxis( Axis a )
**            delete an axis
*/
    public void deleteAxis(Axis a) {

        if (a != null) {
            a.detachAll();
            axis.removeElement(a);
        }
    }
/*
** Vector getMarkerVector(int m)
**         Given the marker index (m > 0 ) return the vector containing the
**         points to stroke the marker
*/
    public Vector getMarkerVector(int m) {

        if (marker == null) return null;


        if (marker.number > 0 && m < marker.number) {

            return marker.vert[m - 1];
        }

        return null;
    }
/*
** boolean loadMarkerFile(URL name)
**        Load the marker file given the files URL. Markers are NOT
** predefined but must be loaded from a Marker definition file
*/
    public boolean loadMarkerFile(URL name) {
        //name = new URL("file://hd/current/dewitt/diffraction%20%B9/marker.txt");
        System.out.println("Loading marker file");
        marker = new Markers(name);

        if (marker.number <= 0) return false;

        return true;

    }
/*
** paint(Graphics g)
**                  Paint it ALL.
*/
    public void paint(Graphics g) {
        Graphics lg = g.create();
        Rectangle r = getBounds();
        if (!paintAll) return;

        r.x += borderLeft;
        r.y += borderTop;
        r.width -= borderLeft + borderRight;
        r.height -= borderBottom + borderTop;



        if (!axis.isEmpty())
            r = drawAxis(lg, r);
        else
            drawFrame(lg, r.x, r.y, r.width, r.height);

        if (imageData == null)
           drawData(lg, r);
        else lg.drawImage(imageData,r.x,r.y,r.width,r.height,this);

        lg.dispose();


    }

    private void drawData(Graphics lg, Rectangle r) {
        int i;
        if (!dataset.isEmpty()) {

            for (i = 0; i < dataset.size(); i++) {
                ((DataSet) dataset.elementAt(i)).draw_legend(lg, r);
            }

            lg.clipRect(r.x, r.y, r.width, r.height);

            datarect.x = r.x;
            datarect.y = r.y;
            datarect.width = r.width;
            datarect.height = r.height;

            for (i = 0; i < dataset.size(); i++) {
                ((DataSet) dataset.elementAt(i)).draw_data(lg, r);
            }
        }
    }



/*
** update(Graphics g)
**                   This method is called through the repaint() method.
** All it does is blank the canvas before calling paint.
*/
    public void update(Graphics g) {
        Rectangle r = getBounds();
        Color c = g.getColor();

//          System.out.println("Graph2d update method called");

        if (clearAll) {
            g.setColor(getBackground());
            g.fillRect(r.x, r.y, r.width, r.height);
            g.setColor(c);
        }
        if (paintAll) paint(g);
    }
/*
** Handle  events. Only one event is handled the pressing
** of the key 'r' - this will repaint the canvas.
*/  public MouseAdapter getMouseResponse() {
    return new MouseAdapter() {
        public void mousePressed(MouseEvent event) {
            repaint();
        }
    };
}



/*
** startedloading()
**                Graph2D will pause until the data is loaded
*/
    public void startedloading() {

        loadingData++;


        if (loadingData != 1) return;

        if (load_thread == null) load_thread = new LoadMessage(this);
        load_thread.setFont(new Font("Helvetica", Font.PLAIN, 25));

        load_thread.begin();

    }
/*
** finishedloading()
**                Graph2D will plot all data if loadingData == 0
*/
    public void finishedloading() {

        loadingData--;

        if (loadingData > 0) return;

        if (load_thread != null) load_thread.end();
        load_thread = null;

    }
/*
** loadmessage(String s)
**               Change the message to display while loading data
*/
    public void loadmessage(String s) {
        if (load_thread == null) load_thread = new LoadMessage(this);
        load_thread.setMessage(s);
    }


    /**
     * *************** * * Protected Methods * *****************
     */
/*
**  Rectangle drawAxis(Graphics g, Rectangle r)
**
**  Draw the Axis. As each axis is drawn and aligned less of the canvas
**  is avaliable to plot the data. The returned Rectangle is the canvas
**  area that the data is plotted in.
*/
    protected Rectangle drawAxis(Graphics g, Rectangle r) {
        Axis a;
        int waxis;
        int x = r.x;
        int y = r.y;
        int width = r.width;
        int height = r.height;

// Calculate the available area for the data
        for (int i = 0; i < axis.size(); i++) {
            a = ((Axis) axis.elementAt(i));

            waxis = a.getAxisWidth(g);

            switch (a.getAxisPos()) {
                case Axis.LEFT:
                    x += waxis;
                    width -= waxis;
                    break;
                case Axis.RIGHT:
                    width -= waxis;
                    break;
                case Axis.TOP:
                    y += waxis;
                    height -= waxis;
                    break;
                case Axis.BOTTOM:
                    height -= waxis;
                    break;
            }
        }
// Draw a frame around the data area (If requested)
        if (frame) drawFrame(g, x, y, width, height);

// Now draw the axis in the order specified aligning them with the final
// data area.
        for (int i = 0; i < axis.size(); i++) {
            a = ((Axis) axis.elementAt(i));

            a.data_window = new Dimension(width, height);

            switch (a.getAxisPos()) {
                case Axis.LEFT:
                    r.x += a.width;
                    r.width -= a.width;
                    a.positionAxis(r.x, r.x, y, y + height);
                    if (r.x == x) {
                        a.gridcolor = gridcolor;
                        a.drawgrid = drawgrid;
                        a.zerocolor = zerocolor;
                        a.drawzero = drawzero;
                    }

                    a.drawAxis(g);

                    a.drawgrid = false;
                    a.drawzero = false;

                    break;
                case Axis.RIGHT:
                    r.width -= a.width;
                    a.positionAxis(r.x + r.width, r.x + r.width, y, y +
                            height);
                    if (r.x + r.width == x + width) {
                        a.gridcolor = gridcolor;
                        a.drawgrid = drawgrid;
                        a.zerocolor = zerocolor;
                        a.drawzero = drawzero;
                    }

                    a.drawAxis(g);

                    a.drawgrid = false;
                    a.drawzero = false;

                    break;
                case Axis.TOP:
                    r.y += a.width;
                    r.height -= a.width;
                    a.positionAxis(x, x + width, r.y, r.y);
                    if (r.y == y) {
                        a.gridcolor = gridcolor;
                        a.drawgrid = drawgrid;
                        a.zerocolor = zerocolor;
                        a.drawzero = drawzero;
                    }

                    a.drawAxis(g);

                    a.drawgrid = false;
                    a.drawzero = false;


                    break;
                case Axis.BOTTOM:
                    r.height -= a.width;
                    a.positionAxis(x, x + width, r.y + r.height, r.y + r.height);
                    if (r.y + r.height == y + height) {
                        a.gridcolor = gridcolor;
                        a.drawgrid = drawgrid;
                        a.zerocolor = zerocolor;
                        a.drawzero = drawzero;
                    }

                    a.drawAxis(g);

                    a.drawgrid = false;
                    a.drawzero = false;


                    break;
            }
        }

        return r;
    }
/*
** drawFrame(Graphics g, int x, int y, int width, int height)
**    If requested draw a frame around the data area.
*/
    protected void drawFrame(Graphics g, int x, int y, int width, int height) {
        g.drawRect(x, y, width, height);
    }


    public Vector getDataset() {
        return dataset;
    }

    public void setDataset(Vector dataset) {
        this.dataset = dataset;
    }

    public boolean isClearAll() {
        return clearAll;
    }

    public void setClearAll(boolean clearAll) {
        this.clearAll = clearAll;
    }

    public boolean isPaintAll() {
        return paintAll;
    }

    public void setPaintAll(boolean paintAll) {
        this.paintAll = paintAll;
    }

    public Image getImageData() {
        return imageData;
    }

    public void setImageData(Image imageData) {
        this.imageData = imageData;
    }
}

/*
** class FileFormatException
** This should be thrown if the fileloader encounters a format error
** not being thrown at the moment. Is being thown by the getMarker method.
*/

class FileFormatException extends Exception {
    public FileFormatException(String s) {
        super(s);
    }
}


/*
** Class LoadMessage
**                   This is a separate thread that flashes a message
** on the canvas that data is loading
*/

class LoadMessage extends Thread {
    private GraphComponent g2d;
    private String message = "Loading Data ... Please Wait!";
    private String newmessage = null;
    private long visible = 500;
    private long invisible = 200;
    private Color foreground = Color.red;
    private Graphics lg = null;
    private Font f = null;


    public LoadMessage(GraphComponent g2d) {
        this.g2d = g2d;

    }

    public LoadMessage(GraphComponent g2d, String s) {

        this(g2d);
        this.message = s;

    }

    public LoadMessage(GraphComponent g, String s, long visible, long invisible) {
        this(g, s);
        this.visible = visible;
        this.invisible = invisible;
    }


    public void begin() {

        g2d.setClearAll(false);
        g2d.setPaintAll(false);

        super.start();

    }

    public void end() {


        g2d.setClearAll(true);
        g2d.setPaintAll(true);

        if (lg != null) lg.dispose();

        g2d.repaint();

    }


    public void run() {
        boolean draw = true;
        FontMetrics fm;
        Rectangle r;
        int sw = 0;
        int sa = 0;
        int x = 0;
        int y = 0;

        setPriority(Thread.MIN_PRIORITY);


        while (true) {

            if (newmessage != null && draw) {
                message = newmessage;
                newmessage = null;
            }

            if (lg == null) {
                lg = g2d.getGraphics();
                if (lg != null) lg = lg.create();
            }

            if (lg != null) {
                if (f != null) lg.setFont(f);
                fm = lg.getFontMetrics(lg.getFont());
                sw = fm.stringWidth(message);
                sa = fm.getAscent();
            } else {
                draw = false;
            }

            if (draw) {
                lg.setColor(foreground);
                r = g2d.getBounds();
                x = r.x + (r.width - sw) / 2;
                y = r.y + (r.height + sa) / 2;
                lg.drawString(message, x, y);

                g2d.repaint();

                try {
                    sleep(visible);
                } catch (Exception e) {
                }
            } else {
                if (lg != null) {
                    lg.setColor(g2d.getBackground());
                    lg.drawString(message, x, y);

                    g2d.repaint();
                }

                try {
                    sleep(invisible);
                } catch (Exception e) {
                }

            }

            draw = !draw;

        }
    }

    public void setFont(Font f) {
        this.f = f;
    }


    public void setMessage(String s) {
        if (s == null) return;
        newmessage = s;
    }


}
